import sys, traceback
import os
import requests as requests
from time import sleep
import simpleeval
from urllib.parse import quote

import termbrowser.adom

class Browser:
	def __init__(self, initialURL: str):
		self.document = termbrowser.adom.Document(self)
		self.URL = initialURL
		self.loading = True
		self.context = {}
		self.debugHistory = "Debugger: Press Alt+K to close.\n"
		self.debugMode = False

	def start_load(self):
		if not self.loading:
			return
		self.evaluator = self.createEvaluator() # reset evaluator before new page is loaded
		self.document = loadFromURL(self.URL, self)
		self.loading = False
		self.document.call_document_action("start", {})
		if self.document.hasInputs:
			self.document.focus_next()
			focused = self.document.get_focused_element()
			if focused.getAttribute("autofocus") == "no":
				self.document.unfocus()

	def createEvaluator(self):
		evaluator = simpleeval.EvalWithCompoundTypes()
		evaluator.functions = simpleeval.DEFAULT_FUNCTIONS
		evaluator.functions.update(self.script_functions())
		return evaluator
	
	# returns if the browser should exit
	def open_link(self, URL: str) -> bool:
		if URL != "term://exit":
			if URL.startswith("term://") and not self.URL.startswith("term://"):
				self.document = termbrowser.adom.Document(self)
				self.document.elements.append(termbrowser.adom.createTextElement("Can not open term:// links. Only Term can open these links."))
			else:
				self.URL = URL
				self.loading = True
			return False
		else:
			return True

	def var(self, name: str, value):
		self.debug(name + ": " + value)
		self.context[name] = value

	def getvar(self, name: str):
		if name in self.context:
			return self.context[name]
		else:
			return "None"

	def encode(self, text: str):
		return quote(text)

	def debug(self, text: str):
		self.debugHistory += text + "\n"
	
	def action(self, name: str):
		self.document.call_action(name, {})

	def script_functions(self):
		return {
			"visit": self.open_link,
			"getvar": self.getvar,
			"var": self.var,
			"action": self.action,
			"encode": self.encode,
			"debug": self.debug
		}

"""
term:// -> files in the ./local/ folder of the term project
file:// -> files on the user's computer
http:// & https:// -> files to curl from the internet
"""
def loadFromURL(URL: str, browser: Browser):
	protocol = None
	protocols = ["term://", "http://", "https://"]
	for p in protocols:
		if URL.startswith(p):
			protocol = p
			break
	if protocol == None:
		return termbrowser.adom.Document(browser).with_message("Error: Could not determine URL protocol `{}`".format(URL))
	elif protocol == "term://":
		file_path = os.path.dirname(os.path.realpath(__file__)) + "/local/" + URL[len("term://"):] + ".term"
		try:
			stream = open(file_path)
			res = stream.read()
			stream.close()
			return termbrowser.adom.term2doc(res, browser)
		except FileNotFoundError:
			return termbrowser.adom.Document(browser).with_message("Could not load `{}`".format(file_path))
	elif protocol == "https://" or protocol == "http://":
		couldOpen = True
		sleep(.1)
		try:
			return makeRequest(browser, URL)
		except Exception as e:
			browser.debug(str(e))
			return termbrowser.adom.Document(browser).with_message("Could not load URL. \n" + str(e))

def makeRequest(browser: Browser, url: str):
	try:
		page = requests.get(url, headers={"content-type": "term"}).text
		return termbrowser.adom.term2doc(page, browser)
	except Exception as e:
		browser.debug(str(e))
		return termbrowser.adom.Document(browser).with_message("Could not load URL. \n" + str(e))
	return termbrowser.adom.Document(browser).with_message("Could not load URL.")


import curses

def get_pieces(styles: list, output_len: int):
	pieces = []
	for i in range(len(styles)):
		current = styles[i]
		follow = None
		if i < len(styles) - 1:
			follow = styles[i + 1]

		startPos = current.start

		endPos = output_len
		if follow != None:
			endPos = follow.start

		col = curses.A_REVERSE

		if current.style == "bold":
			col += curses.A_BOLD
		elif current.style == "underline":
			col += curses.A_UNDERLINE
		pieces.append((startPos, endPos, col))
	return pieces


from typing import List
import simpleeval

import termbrowser.browser

class Action:
	def __init__(self, name, contents):
		self.name = name
		self.code = contents

class Attribute:
	def __init__(self, name, value):
		self.name = name
		self.value = value

class Element:
	def __init__(self, type):
		self.type = type
		self.value = None
		self.attributes = []
		self.children = []
		self.focused = False
		self.focus_cursor_index = 0
	
	def setAttribute(self, name, value):
		self.attributes.append(Attribute(name, value))

	def getAttribute(self, name) -> str:
		if name != "":
			for item in self.attributes:
				if item.name == name:
					return item.value
			return None
		else:
			return None

	def appendChild(self, child):
		self.children.append(child)

def createTextElement(value):
	t = Element("text")
	t.value = value
	return t
	

class DocumentLink:
	def __init__(self, key: chr, URL: str):
		self.key = key
		self.URL = URL

class Document:
	def __init__(self, browser: termbrowser.browser.Browser):
		self.browser = browser
		self.links = []
		self.elements = []
		self.actions = []
		self.focus = -1
		self.hasInputs = False
		self.evaluator = self.createEvaluator()

	def createEvaluator(self):
		e = simpleeval.EvalWithCompoundTypes()
		e.functions = simpleeval.DEFAULT_FUNCTIONS
		e.functions.update(self.browser.script_functions())
		return e
	
	def with_message(self, message: str):
		self.elements = [createTextElement(message)]
		return self

	def focus_next(self):
		self.focus = self._search_focus(self.focus)
		focused = self.get_focused_element()
		focused.focus_cursor_index = len(focused.value)

	def get_focused_element(self):
		return self.get_focus_list()[self.focus]

	def unfocus(self):
		focus_list: List[Element] = get_all_elements(self.elements)
		for item in focus_list:
			item.focused = False
		self.focus = -1

	def get_focus_list(self):
		all: List[Element] = get_all_elements(self.elements)
		focusable = ("input",)
		focusList: List[Element] = []
		for element in all:
			if element.type.startswith(focusable):
				focusList.append(element)
		return focusList
	
	def _search_focus(self, start: int):
		all: List[Element] = get_all_elements(self.elements)
		focusable = ("input",)
		focusList: List[Element] = []
		for element in all:
			element.focused = False
			if element.type.startswith(focusable):
				focusList.append(element)
		index = start + 1
		index = index if index < len(focusList) else 0
		if len(focusList) > 0:
			focusList[index].focused = True
			return index
		else:
			return -1

	def add_link(self, key: str, URL: str):
		self.links.append(DocumentLink(ord(key), URL))

	def find_link(self, key: int):
		index = 0
		for link in self.links:
			if link.key == key:
				return index
			index += 1
		return -1

	def find_action(self, name: str) -> Action:
		for a in self.actions:
			if a.name == name:
				return a
		return None

	def submit(self, element: Element):
		self.call_element_action(element, "submit", {
			"value": element.value
		})
	
	def change(self, element: Element):
		self.call_element_action(element, "change", {
			"value": element.value
		})

	def call_action(self, name: str, args: dict):
		action = self.find_action(name)
		if action == None:
			return
		self.evaluator.names = args
		self.evaluator.eval("(" + action.code + ")")

	def call_element_action(self, element: Element, name: str, args: dict):
		actionCall = element.getAttribute(name)
		if actionCall == None:
			return
		action = self.find_action(actionCall)
		if action == None:
			return
		self.evaluator.names = args
		self.evaluator.eval("(" + action.code + ")")

	def call_document_action(self, name: str, args: dict):
		if name == None:
			return
		action = self.find_action("[" + name + "]")
		if action == None:
			return
		self.evaluator.names = args
		self.evaluator.eval("(" + action.code + ")")

		

def get_all_elements(elementList: List[Element]):
	res = []
	for element in elementList:
		res.append(element)
		if len(element.children) > 0:
			res.extend(get_all_elements(element.children))
	return res

elementTypes = [
	"cont",
	"text",
	"link",
	"input",
	"br"
]

validFormats = [
	"m100"
]

def _crashDoc(reason: str, lineNumber: int, browser):
	doc = Document(browser)
	doc.elements = [createTextElement("Line {}: ".format(lineNumber + 1) + reason)]
	return doc

def term2doc(contents, browser):
	res = Document(browser)

	contents = contents.strip(" \n\r")
	lines: list = contents.splitlines()
	
	elementQueue: list = []

	readTo = -1
	for lineNumber in range(len(lines)):
		command: str = lines[lineNumber].strip()

		if lineNumber == 0:
			if not command.startswith("@termtype"):
				return _crashDoc("Term files must start with @termtype:[a valid termtype].\nex. @termtype:m100", lineNumber, browser)
			else:
				termstart = digestAttribute(command[1:])
				passes = False
				for tt in validFormats:
					if termstart["value"] == tt:
						passes = True
						break
				if not passes:
					return _crashDoc("Term file is not in a valid format.", lineNumber, browser)
				continue
				
		if readTo > lineNumber:
			continue

		top: Element = getTop(elementQueue)

		if command.startswith("#"):
			continue
		elif command.startswith("-"):
			if top == None:
				return _crashDoc("No element for attribute to be assigned.", lineNumber, browser)
			else:
				result: dict = digestAttribute(command[1:])

				if result == None:
					return _crashDoc("Could not interpret attribute.", lineNumber, browser)
				else:
					top.setAttribute(result["name"], result["value"])
					if result["name"] == "initial" and top.type == "input":
						top.value = result["value"]
			
		elif command.startswith(tuple(elementTypes)):
			etype, evalue = digestDeclaration(command)
			newElement = Element(etype)
			newElement.value = evalue

			if top != None:
				top.children.append(newElement)

			if etype == "input":
				res.hasInputs = True
			
			if etype != "br": # br does not have an ending tag and cannot have chilren
				elementQueue.append(newElement)
			
		elif command.startswith("end"):
			if top == None:
				return _crashDoc("No element for `end` keyword to complete.", lineNumber, browser)
			else:
				ele: Element = elementQueue.pop()

				if ele.type == "link":
					linkKey = ele.getAttribute("key")
					linkURL = ele.getAttribute("url")
					if(linkKey == None or linkURL == None):
						return _crashDoc("Link element requires attributes (key, url)", lineNumber, browser)
					res.add_link(linkKey, linkURL)

				if len(elementQueue) == 0:
					res.elements.append(ele)
		elif command.startswith("action"):
			comIndex = command.find(":")
			if comIndex == -1:
				return _crashDoc("Could not find `:` following the `action` declaration. This needs to be on the same line.", lineNumber, browser)
			between = command[len("action"):comIndex]
			if between.strip() != "":
				return _crashDoc("Unrecognized `" + between + "`.", lineNumber, browser)
			
			# determine action name
			startIndex = command.find("(", comIndex)
			if startIndex == -1:
				return _crashDoc("Could not find `(` following the `action:` declaration. This needs to be on the same line.", lineNumber, browser)

			actionName = command[comIndex + 1:startIndex].strip()

			# track opening parenthesis to determine the end of the action, set to 1 because the action block is wrapped in parenthesis
			openCount = 1

			firstRow = command[startIndex + 1:].strip()

			# start to build the contents of the action
			actionContents = firstRow + "\n"

			for c in firstRow:
				if c == "(":
					openCount += 1
				elif c == ")":
					openCount -= 1

			# move onto other lines, action block is still open
			while openCount != 0 and lineNumber < len(lines) - 1:
				lineNumber += 1
				newLine = lines[lineNumber].strip()
				for c in newLine:
					if c == "(":
						openCount += 1
					elif c == ")":
						openCount -= 1

					if openCount < 0:
						return _crashDoc("Unexpected `)`.", lineNumber, browser)
				
				actionContents += newLine + "\n"

			if openCount > 0:
				return _crashDoc("Expected ending `)` for action block.", lineNumber, browser)
			elif openCount < 0:
				return _crashDoc("Unexpected `)`.", lineNumber, browser)
			
			actionContents = actionContents[:actionContents.rfind(")")]

			readTo = lineNumber + 1

			res.actions.append(Action(actionName, actionContents))
		elif len(command) > 0:
			return _crashDoc("Could not interpret line format.", lineNumber, browser)
	return res

def digestDeclaration(line: str) -> tuple:
	restype: str = None
	resvalue: str = ""
	for etype in elementTypes:
		if line.startswith(etype):
			restype = etype
			break
	
	if restype == None:
		return None
	else:
		colIndex = line.find(":")
		if colIndex != -1:
			resvalue = line[colIndex + 1:]
		return (restype, resvalue)

def getTop(queue) -> Element:
	return queue[-1] if len(queue) > 0 else None

def digestAttribute(line) -> dict:
	colIndex = line.find(":")
	if colIndex == -1:
		return None
	
	attribName: str = line[:colIndex]
	if len(attribName) == 0:
		return None
	else:
		attribValue: str = line[colIndex + 1:]
		if len(attribValue) == 0:
			return None
		else:
			return { "name": attribName, "value": attribValue.strip() }


from ..adom import Document, Element
from ..vector import *
from ..util import *
from .borders import getBorderCodes
from .style import OutputStyle

from textwrap import TextWrapper

text_wrapper = TextWrapper(
	replace_whitespace=False,
	break_long_words=True,
	expand_tabs=True,
	tabsize=4,
	#drop_whitespace=True # try this out??
)

# all possible text styling options
TextStyles = (
	"bold",
	"underline"
)

DEFAULT_INPUT_WIDTH = 15

# document --(render)--> list of rows (strings)
def renderDocument(document: Document, width: int, height: int, scroll: int):
	# initialize frame with cleared screen
	res = clearScreen(width, height)

	# initialize frame with single normal style
	styles: List[OutputStyle] = [
		OutputStyle(Vec(0, 0), "normal")
	]

	# cursor starts at the top left corner
	cursor = Vec(0, 0)

	# recursively render all elements to the screen
	for element in document.elements:
		writeSize = renderElement(element, cursor.x, cursor.y - scroll, width, height, res, styles, None)
		# move the cursor down using the size of the rendered element
		cursor.y += writeSize.y

	# combine rows
	out = ""
	for row in res:
		out += row

	# style start changes from Vec(x, y) -> out[i]
	for style in styles:
		style.start = style.start.x + (style.start.y * width)

	return (out, styles)

# row list filled with empty
def clearScreen(width: int, height: int) -> list:
	return [" " * width] * height

def renderDebugger(text: str, width: int, height: int) -> str:
	textlines = text.splitlines()
	if len(textlines) > height:
		textlines = textlines[-height:]
	out = ""
	for line in textlines:
		out += restrict_len(expand_len(line, width), width)

	for i in range(height - len(textlines)):
		out += " " * width

	return out

def renderElement(element: Element, x: int, y: int, WIDTH: int, HEIGHT: int, res: list, styles: list, parentSize: Vec):
	# the write size is used to determine how far away the next element in queue should be placed
	writeSize = Vec(0, 0)

	if element.type == "cont":
		padding = getPadding(element, WIDTH, HEIGHT)
		direction = getDirection(element)
		borderType = element.getAttribute("border")
		paddingSize = Vec(padding['left'] + padding['right'], padding['top'] + padding['bottom'])
		initialOffset = Vec(padding['left'], padding['top'])
		defSize = getDefinedSize(element, WIDTH, HEIGHT)

		hasDefWidth = defSize.x != -1
		hasDefHeight = defSize.y != -1

		if borderType != None:
			initialOffset.add(1, 1)

		offset = cloneVec(initialOffset)

		childrenSize = Vec(0, 0)

		# get children size (uncalculated pc)
		for child in element.children:
			defSingleSize = getDefinedSize(child, WIDTH, HEIGHT)
			otherSingleSize = getElementSize(child, WIDTH, HEIGHT)
			singleSize = Vec(
				otherSingleSize.x if defSingleSize.x == -1 else defSingleSize.x,
				otherSingleSize.y if defSingleSize.y == -1 else defSingleSize.y
			)

			if direction == "row":
				childrenSize.x += singleSize.x
				childrenSize.y = max(childrenSize.y, singleSize.y)
			elif direction == "column":
				childrenSize.y += singleSize.y
				childrenSize.x = max(childrenSize.x, singleSize.x)

		# get defined size -> fallback on children collective size
		contSize = Vec(
			defSize.x if hasDefWidth else childrenSize.x + paddingSize.x,
			defSize.y if hasDefHeight else childrenSize.y + paddingSize.y
		)

		# enforce box-sizing
		if borderType != None:
			contSize.add(-2 if hasDefWidth else 0, -2 if hasDefHeight else 0)

		# render children with clamp, calculated pc
		for child in element.children:
			singleSize = renderElement(child, x + offset.x, y + offset.y, WIDTH, HEIGHT, res, styles, contSize)

			if direction == "row":
				offset.x += singleSize.x
			elif direction == "column":
				offset.y += singleSize.y

		# render border
		if borderType != None:
			renderBorder(borderType, Vec(x, y), contSize, res)
			writeSize = Vec(contSize.x + 2, contSize.y + 2)
		else:
			writeSize = contSize

	elif element.type == "text":
		# has value
		if element.value != "":
			outerSize = parentSize if parentSize != None else Vec(WIDTH, HEIGHT)
			alignOffset = getAlignOffset(element, outerSize)
			maxWidth = outerSize.x
			widthAttr = element.getAttribute("width")
			renderWidth = parseSize(widthAttr, maxWidth) if widthAttr != None else maxWidth

			wrapped_text_size = getWrapAndSize(element.value, renderWidth)
			wrapped_text = wrapped_text_size["text"]
			wrapped_size = wrapped_text_size["size"]

			writeSize = wrapped_size
			startPos = x + alignOffset

			textStyle = element.getAttribute("style")

			renderRows = wrapped_text.splitlines()
			for rowIndex in range(len(renderRows)):
				rowText = renderRows[rowIndex]
				rowTextLen = len(rowText)
				renderY = y + rowIndex

				# in bounds
				if renderY < len(res) and renderY >= 0:
					# styled text
					if textStyle != None and textStyle.startswith(TextStyles):
						styles.append(OutputStyle(Vec(startPos, renderY), textStyle))
						styles.append(OutputStyle(Vec(startPos + rowTextLen, renderY), "normal"))
					# render single line of text
					res[renderY] = res[renderY][0:startPos] + rowText + res[renderY][startPos + rowTextLen:]
		else:
			writeSize = Vec(0, 1)

	elif element.type == "link":
		toRender = getLinkText(element)
		toRenderLength = len(toRender)

		outerSize = parentSize if parentSize != None else Vec(WIDTH, HEIGHT)
		alignOffset = getAlignOffset(element, outerSize)
		maxWidth = WIDTH
		if parentSize != None:
			maxWidth = parentSize.x
		widthAttr = element.getAttribute("width")
		renderWidth = parseSize(widthAttr, maxWidth) if widthAttr != None else maxWidth

		wrapped = getWrapAndSize(toRender, renderWidth)

		writeSize = wrapped["size"]
		startPos = x + alignOffset

		textStyle = element.getAttribute("style")

		renderRows = wrapped["text"].splitlines()
		for rowIndex in range(len(renderRows)):
			rowText = renderRows[rowIndex]
			rowTextLen = len(rowText)
			renderY = y + rowIndex

			if renderY < len(res) and renderY >= 0:
				endPos = startPos + rowTextLen
				if textStyle != None and textStyle.startswith(TextStyles):
					styles.append(OutputStyle(Vec(startPos, renderY), textStyle))
					styles.append(OutputStyle(Vec(endPos, renderY), "normal"))
				res[renderY] = res[renderY][0:startPos] + rowText + res[renderY][endPos:]

	elif element.type == "input":
		defSize = getDefinedSize(element, WIDTH, HEIGHT)
		calcWidth = defSize.x if defSize.x != -1 else DEFAULT_INPUT_WIDTH
		renderCursor = "\N{FULL BLOCK}" if element.focused else ""
		val = element.value
		idx = element.focus_cursor_index
		draw_cursor_end = val[idx + 1:] if idx < len(val) and idx != -1 else ""
		draw_cursor = val[:idx] + renderCursor + draw_cursor_end
		toRender = rcaplen(
			expand_len(
				draw_cursor if idx != -1 else val,
				calcWidth
			),
			calcWidth
		)

		toRenderLength = len(toRender)

		outerSize = parentSize if parentSize != None else Vec(WIDTH, HEIGHT)
		alignOffset = getAlignOffset(element, outerSize)

		borderType = "dotted thick" if element.focused else "dotted thin"
		renderBorder(borderType, Vec(x, y), Vec(toRenderLength, 1), res)
		writeSize = Vec(toRenderLength + 2, 3)
		startPos = x + alignOffset + 1
		renderY = y + 1
		if renderY < len(res) and renderY >= 0:
			res[renderY] = res[renderY][0:startPos] + toRender + res[renderY][startPos + toRenderLength:]

	elif element.type == "br":
		writeSize = Vec(1, 1)

	return writeSize

def getLinkText(element):
	return "[" + element.getAttribute("key") + "] " + element.value

def getWrapAndSize(text: str, maxWidth: int):
	if maxWidth == None:
		return {
			"text": text,
			"size": Vec(len(text), 1)
		}
	text_wrapper.width = maxWidth
	wrappedText = text_wrapper.fill(text)
	return {
		"text": wrappedText,
		"size": Vec(maxWidth if maxWidth < len(text) else len(text), len(wrappedText.splitlines()))
	}

def getAlignOffset(element: Element, parentSize: Vec) -> int:
	val = element.value
	if element.type == "link":
		val = getLinkText(element)

	align = element.getAttribute("align")
	defWidth = getDefinedSize(element, parentSize.x, parentSize.y).x
	wrapped = getWrapAndSize(val, defWidth if defWidth != -1 else parentSize.x)
	if align == "center":
		return round(float(parentSize.x) / 2.0) - round(float(wrapped["size"].x) / 2.0)
	return 0

def getElementSize(element: Element, WIDTH: int, HEIGHT: int) -> Vec:
	if element.type == "text":
		widthAttr = element.getAttribute("width")
		renderWidth = parseSize(widthAttr, WIDTH) if widthAttr != None else WIDTH
		wrapped_text_size = getWrapAndSize(element.value, renderWidth)
		return wrapped_text_size["size"]
	elif element.type == "link":
		widthAttr = element.getAttribute("width")
		renderWidth = parseSize(widthAttr, WIDTH) if widthAttr != None else WIDTH
		wrapped_text_size = getWrapAndSize(getLinkText(element), renderWidth)
		return wrapped_text_size["size"]
	elif element.type == "input":
		widthAttr = element.getAttribute("width")
		renderWidth = parseSize(widthAttr, WIDTH) if widthAttr != None else WIDTH
		return Vec(renderWidth + 2, 3)
	elif element.type == "cont":
		childrenSize = Vec(0, 0)
		padding = getPadding(element, WIDTH, HEIGHT)
		paddingSize = Vec(padding['left'] + padding['right'], padding['top'] + padding['bottom'])
		defSize = getDefinedSize(element, WIDTH, HEIGHT)
		hasDefWidth = defSize.x != -1
		hasDefHeight = defSize.y != -1

		for child in element.children:
			singleSize = getElementSize(child, WIDTH, HEIGHT)
			direction = getDirection(element)
			if direction == "row":
				childrenSize.x += singleSize.x
				childrenSize.y = max(childrenSize.y, singleSize.y)
			elif direction == "column":
				childrenSize.y += singleSize.y
				childrenSize.x = max(childrenSize.x, singleSize.x)

		res = Vec(
			defSize.x if hasDefWidth else childrenSize.x + paddingSize.x,
			defSize.y if hasDefHeight else childrenSize.y + paddingSize.y
		)

		return res
	elif element.type == "br":
		return Vec(1, 1)
	return None

def renderBorder(type: str, pos: Vec, size: Vec, res: list):
	screenSize = Vec(len(res[0]), len(res))
	borderCodes = getBorderCodes(type)

	# any of border visible
	if pos.y >= len(res):
		return

	# first row
	if pos.y >= 0:
		firstRow = ""

		# margin
		firstRow += res[pos.y][0:pos.x]

		# top left corner
		firstRow += borderCodes["top-left"]
		# top
		firstRow += borderCodes["top"] * size.x
		# top right corner
		firstRow += borderCodes["top-right"]

		firstRow += res[pos.y][pos.x + 1 + size.x + 1:screenSize.x]

		# establish first row
		res[pos.y] = firstRow if len(firstRow) <= screenSize.x else firstRow[:screenSize.x]

	# middle rows
	for y in range(size.y):
		borderContentY = pos.y + 1 + y
		if borderContentY < len(res) and borderContentY >= 0:
			newRow = res[borderContentY][0:pos.x]
			newRow += borderCodes["left"]
			newRow += res[borderContentY][pos.x + 1:pos.x + 1 + size.x]
			newRow += borderCodes["right"]
			newRow += res[borderContentY][pos.x + 1 + size.x + 1:screenSize.x]
			res[borderContentY] = newRow if len(newRow) <= screenSize.x else newRow[:screenSize.x]

	# last row
	lastBorderContentY = pos.y + size.y + 1
	if lastBorderContentY < len(res) and lastBorderContentY >= 0:
		lastRow = ""
		# margin
		lastRow += res[lastBorderContentY][0:pos.x]

		# bottom left corner
		lastRow += borderCodes["bottom-left"]
		# bottom
		lastRow += borderCodes["bottom"] * size.x
		# bottom right corner
		lastRow += borderCodes["bottom-right"]

		lastRow += res[lastBorderContentY][pos.x + 1 + size.x + 1:screenSize.x]

		# establish last row
		res[lastBorderContentY] = lastRow if len(lastRow) <= screenSize.x else lastRow[:screenSize.x]

def getDefinedSize(element: Element, WIDTH: int, HEIGHT: int) -> Vec:
	res = Vec(-1, -1)
	defWidth = element.getAttribute("width")
	defHeight = element.getAttribute("height")
	padding = getPadding(element, WIDTH, HEIGHT)
	if defWidth != None:
		res.x = parseSize(defWidth, WIDTH)
	if defHeight != None:
		res.y = parseSize(defHeight, HEIGHT)
	return res

def parseSize(size, MAX) -> int:
	if isInt(size):
		return int(size)
	else:
		last = size[-2:]
		start = size[:-2]
		if isInt(start):
			value = int(start)
			if last == "pc":
				return round(float(MAX) / 100.0 * value)
			else:
				return -1
		else:
			return -1

def getPadding(element: Element, width: int, height: int) -> dict:
	paddingOut = {
		'top': 0,
		'bottom': 0,
		'left': 0,
		'right': 0
	}

	padding = element.getAttribute("padding")
	padding_top = element.getAttribute("padding-top")
	padding_bottom = element.getAttribute("padding-bottom")
	padding_left = element.getAttribute("padding-left")
	padding_right = element.getAttribute("padding-right")

	if padding != None:
		vert = parseSize(padding, height)
		horiz = parseSize(padding, width)
		paddingOut = {
			'top': vert,
			'bottom': vert,
			'left': horiz,
			'right': horiz
		}

	if padding_top != None:
		paddingOut['top'] += parseSize(padding_top, height)

	if padding_bottom != None:
		paddingOut['bottom'] += parseSize(padding_bottom, height)

	if padding_left != None:
		paddingOut['left'] += parseSize(padding_left, width)

	if padding_right != None:
		paddingOut['right'] += parseSize(padding_right, width)

	return paddingOut

def getDirection(element: Element) -> str:
	dir = element.getAttribute("direction")
	if dir == None:
		return "column"
	else:
		return dir



# > Term Browser <

# interpret args
import argparse
parser = argparse.ArgumentParser()
parser.add_argument("--url", help="the URL you would like to visit")
args = parser.parse_args()
INIT_URL = args.url

import os
import traceback
import curses
import threading
import time
import pyperclip

from .adom import Document, Element
from .browser import *
from .rendering.render import *
from .util import *
from .window import Window
from .pieces import get_pieces

window: Window = Window(curses.initscr())

browser: Browser = Browser(INIT_URL if INIT_URL != None else "term://welcome")

user_input = None
user_input_thread_handle = None
cursor_index: int = -1
scroll: int = 0

paste_mode: bool = False

fps: int = 60 # 20 frames per second

def setup(screen):
	window = Window(screen)
	lifecycle()

def lifecycle():
	global window, user_input_thread_handle

	user_input_thread_handle = threading.Thread(name="daemon", target=user_input_thread, args=())
	user_input_thread_handle.setDaemon(True)
	user_input_thread_handle.start()

	render()

	while True:
		try:
			resized = window.get_resized()
			if resized:
				window.resize()
				continue
			update()
			render()
			if browser.loading:
				browser.start_load()
			window.refresh()
			time.sleep(1.0 / fps)
		except KeyboardInterrupt:
			curses.endwin()
			sys.exit(0)
		except Exception as e:
			curses.endwin()
			traceback.print_exc(file=sys.stdout)
			browser.debug(str(e))


def user_input_thread():
	global window, user_input, cursor_index, scroll
	while True:
		user_input = window.get_input()
		linkIndex = browser.document.find_link(user_input)
		if user_input == ord("`") and browser.document.focus == -1:
			window.exiting = True
			curses.nocbreak()
			curses.echo()
			window.screen.keypad(0)
			curses.endwin()
			exit(0)
		elif user_input == 203: # Alt + K (or Alt + I + I?)
			browser.debugMode = not browser.debugMode
		elif user_input == 9: # tab:
			browser.document.focus_next()
			_focused = browser.document.get_focused_element()
		elif browser.document.focus != -1:
			char = chr(user_input)
			inputChars = " abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890!@#$%&*()-=_+[]{}\|'\";:.>,</?`~"
			# use for debugging
			# browser.URL += str(user_input) + char
			if browser.document.focus == -2: # URL
				if user_input == 226: # alt + v to enable paste
					toInsert = remove_spacing(pyperclip.paste())
					browser.URL = browser.URL[:cursor_index] + toInsert + browser.URL[cursor_index:]
					cursor_index += len(toInsert)
				elif user_input == 260: # left arrow
					cursor_index -= 1 if cursor_index != 0 else 0
				elif user_input == 261: # right arrow
					cursor_index += 1 if cursor_index != len(browser.URL) else 0
				elif user_input == 127: # backspace
					if len(browser.URL) == 0 or cursor_index == 0:
						continue
					browser.URL = browser.URL[:cursor_index - 1] + browser.URL[cursor_index:]
					cursor_index -= 1
				elif user_input == 10: # enter
					browser.open_link(browser.URL)
					browser.document.focus = -1
					cursor_index = -1
				elif inputChars.find(char) != -1:
					browser.URL = browser.URL[:cursor_index] + str(char) + browser.URL[cursor_index:]
					cursor_index += 1
			else: # input element
				focusedElement = browser.document.get_focused_element()

				old_index = focusedElement.focus_cursor_index

				if user_input == 226: # alt + v to enable paste
					toInsert = remove_spacing(pyperclip.paste())
					focusedElement.value = focusedElement.value[:old_index] + toInsert + focusedElement.value[old_index:]
					focusedElement.focus_cursor_index += len(toInsert)
				elif user_input == 260: # left arrow
					focusedElement.focus_cursor_index -= 1 if old_index != 0 else 0
				elif user_input == 261: # right arrow
					focusedElement.focus_cursor_index += 1 if old_index != len(focusedElement.value) else 0

				elif inputChars.find(char) != -1:
					focusedElement.value = focusedElement.value[:old_index] + str(char) + focusedElement.value[old_index:]
					focusedElement.focus_cursor_index += 1
					browser.document.change(focusedElement)
				elif user_input == 127: # backspace
					if len(focusedElement.value) == 0 or old_index == 0:
						continue
					focusedElement.value = focusedElement.value[:old_index - 1] + focusedElement.value[old_index:]
					focusedElement.focus_cursor_index -= 1
				elif user_input == 10: # enter
					browser.document.submit(focusedElement)
				elif user_input == 197: # Alt + Q
					browser.document.unfocus()
		elif user_input == 27: # esc
			browser.document.unfocus()
			browser.document.focus = -2
			cursor_index = len(browser.URL)
		if user_input == 259: # up arrow
			scroll -= 1 if scroll > 0 else 0
		elif user_input == 258: # down arrow
			scroll += 1
		elif linkIndex != -1:
			exiting = browser.open_link(browser.document.links[linkIndex].URL)

def update():
	global window, user_input
	if window.exiting:
		exit(0)
	if browser.loading:
		cursor_index = -1
	user_input = ""

def remove_spacing(string: str):
	return string.replace("\t", "").replace("\n", "")

def render():
	window.start_render(0, 0)

	# URL
	url_cursor = "\N{FULL BLOCK}"
	render_url_length = window.WIDTH - 1
	draw_cursor_end = browser.URL[cursor_index + 1:] if cursor_index < len(browser.URL) and cursor_index != -1 else ""
	draw_cursor = browser.URL[:cursor_index] + url_cursor + draw_cursor_end
	render_url = expand_len(
		restrict_len(
			draw_cursor if browser.document.focus == -2 else browser.URL,
			render_url_length
		),
		render_url_length
	)
	window.render(render_url, curses.A_UNDERLINE)
	# Check and Loading symbols
	window.render("\N{HEAVY CHECK MARK}" if not browser.loading else "\N{DOTTED CIRCLE}", curses.A_UNDERLINE)

	# Document
	(output, styles) = renderDocument(browser.document, window.WIDTH, window.HEIGHT, scroll)
	output = restrict_len(
		remove_spacing(output),
		window.WIDTH * (window.HEIGHT - 1) - 1 # remove last line to account for URL bar, remove last character to stop scroll
	)
	render_pieces = get_pieces(styles, len(output))
	window.start_render(1, 0)
	for piece in render_pieces:
		(startPos, endPos, col) = piece
		window.render(output[startPos:endPos], col)

	# Debugger
	debugHeight = 8
	if browser.debugMode:
		window.start_render(window.HEIGHT - debugHeight, 0)
		debugged = renderDebugger(browser.debugHistory, window.WIDTH, debugHeight)[:-1]
		window.render(debugged, curses.A_NORMAL)

	# Disable Cursor
	window.disable_cursor()



curses.wrapper(setup)



def getBorderCodes(type: str) -> dict:
	# defaults are for 'line'
	codes = {
		"top-left": "\u2554",
		"top-right": "\u2557",
		"bottom-left": "\u255A",
		"bottom-right": "\u255D",
		"top": "\u2550",
		"bottom": "\u2550",
		"left": "\u2551",
		"right": "\u2551",
	}

	if type == "dotted thick":
		codes["top-left"] = "\u250F"
		codes["top-right"] = "\u2513"
		codes["bottom-left"] = "\u2517"
		codes["bottom-right"] = "\u251B"
		codes["top"] = "\u2505"
		codes["bottom"] = "\u2505"
		codes["left"] = "\u2507"
		codes["right"] = "\u2507"
	
	elif type == "dotted thin":
		codes["top-left"] = "\u250C"
		codes["top-right"] = "\u2510"
		codes["bottom-left"] = "\u2514"
		codes["bottom-right"] = "\u2518"
		codes["top"] = "\u2504"
		codes["bottom"] = "\u2504"
		codes["left"] = "\u2506"
		codes["right"] = "\u2506"

	return codes